from django.core import validators
from django.db import models
from django.utils.translation import ugettext_lazy as _


class HexColorValidator(validators.RegexValidator):
    """
    Hex color validator (seven-character hexadecimal notation, e.g.: "#0055ff")
    """
    regex = r'^#[0-9a-fA-F]{6}$'
    message = _('Enter a valid seven-character hexadecimal color value.')


class ColorModelField(models.CharField):
    """
    Hex color model field, e.g.: "#0055ff" (It's not a html color picker widget)

    Note: It's not a html color picker, because "<input type="color">" doesn't allow
    empty values! It will always fallback to "#000000"

    If you need a fancy color picker use e.g.: "django-colorfield" ;)
    """
    description = _('Color field (seven-character hexadecimal notation) e.g.: "#0055ff"')

    def __init__(self, **kwargs):
        kwargs['max_length'] = 7
        super().__init__(**kwargs)
        self.validators.append(validators.MinLengthValidator(self.max_length))
        self.validators.append(HexColorValidator())



from django import forms
from django.http import QueryDict

from bx_django_utils.approve_workflow.constants import POST_KEY_APPROVE


class PublishAdminForm(forms.ModelForm):
    """
    Activate models REQUIRED_FIELDS_PUBLIC on approve
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This class used as normal admin ModelForm and for admin inline ModelForm
        # Both will be initialized in sightly other way. So we must do this:
        if 'data' in kwargs:
            # Inline Formset instance made with: form(**defaults)
            post_data = kwargs['data']
        elif args:
            # normal admin ModelForm instance made:
            #   ModelForm(request.POST, request.FILES, instance=obj)
            # or
            #   ModelForm(instance=obj)
            post_data = args[0]
        else:
            return

        assert isinstance(post_data, QueryDict)
        if post_data:
            self.set_required_fields(post_data)

    def get_required_fields(self, post_data):
        """
        Maybe overwrite this method and set other fields depend on post_data
        """
        return self.instance.REQUIRED_FIELDS_PUBLIC

    def set_required_fields(self, post_data):
        if POST_KEY_APPROVE in post_data:
            # "Approve" submit button was used
            # set more field as "required" as in model defined.
            required_fields = self.get_required_fields(post_data)
            for field_name, field in self.fields.items():
                if field_name in required_fields:
                    field.required = True



from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from stdnum import isbn
from stdnum.ean import calc_check_digit, compact
from stdnum.exceptions import InvalidChecksum, InvalidFormat, InvalidLength
from stdnum.exceptions import ValidationError as StdnumValidationError
from stdnum.util import isdigits

from bx_django_utils.data_types.gtin.constants import DEFAULT_ACCEPT_LENGTH


def validate_gtin(number, accepted_length=DEFAULT_ACCEPT_LENGTH):
    """
    It's the same as stdnum.ean.validate() but also accept ISBN-10
    """
    number = compact(number)  # Convert to the minimal representation
    if not isdigits(number):
        raise InvalidFormat()

    if len(number) not in accepted_length:
        raise InvalidLength()

    if len(number) == 10:
        # Maybe a ISBN-10 number:
        return isbn.validate(number)

    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()

    return number


@deconstructible
class GtinValidator:
    """ Validate GTIN number """

    code = 'invalid'

    def __init__(self, accepted_length=DEFAULT_ACCEPT_LENGTH):
        self.accepted_length = accepted_length

    def __call__(self, value):
        try:
            validate_gtin(value, accepted_length=self.accepted_length)
        except StdnumValidationError as err:
            raise ValidationError(err.message, code=self.code)

    def __eq__(self, other):
        return (
            isinstance(other, GtinValidator) and
            self.accepted_length == other.accepted_length
        )



import uuid

from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission


def filter_permission_names(permissions):
    """
    Generate a Permission model query filtered by names, e.g.: ['<app_label>.<codename>', ...]
    """
    permission_pks = []
    for permission in permissions:
        assert '.' in permission, f'Wrong format: {permission!r}'
        app_label, codename = permission.split('.')
        try:
            permission_pk = Permission.objects.only('pk').get(
                codename=codename,
                content_type__app_label=app_label
            ).pk
        except Permission.DoesNotExist as err:
            raise Permission.DoesNotExist(f"{err}: '{app_label}.{codename}'")

        permission_pks.append(permission_pk)
    qs = Permission.objects.filter(pk__in=permission_pks)
    return qs


def assert_permissions(user, permissions):
    """
    Check user permissions.
    This compares all user permissions: From user group and user object.
    """
    all_user_permissions = set()
    for backend in auth.get_backends():
        if hasattr(backend, "get_all_permissions"):
            all_user_permissions.update(backend.get_all_permissions(user))

    expected_permissions = set(permissions)
    assert all_user_permissions == expected_permissions, (
        f"{tuple(sorted(all_user_permissions))!r}"
        ' is not'
        f" {tuple(sorted(expected_permissions))}!r"
    )


def make_test_user(
    username=None,
    email='test@test.tld',
    password='t',
    is_staff=True,
    permissions=None
):
    """
    Create a test user and set given permissions.

    :param permissions: Queryset or list of Permission instances
    :return: User instance
    """
    if username is None:
        # Generate a unique username for tests that will create more than one user
        username = f'test {uuid.uuid4()}'

    for permission in permissions:
        assert isinstance(permission, Permission)

    User = get_user_model()
    user = User.objects.create_user(
        username=username,
        email=email,
        password=password,
        is_staff=is_staff
    )
    user.user_permissions.set(permissions)
    user = User.objects.get(pk=user.pk)
    return user


def make_minimal_test_user(
    username=None,
    email='test@test.tld',
    password='t',
    is_staff=True,
    permissions=None
):
    """
    Create a test user and set given permissions.

    :param permissions: List of strings like: 'auth.import', 'app_label.codename', ...
    :return: User instance
    """
    if permissions is not None:
        # convert string list to queryset
        permissions = filter_permission_names(permissions)

    user = make_test_user(
        username=username,
        email=email,
        password=password,
        is_staff=is_staff,
        permissions=permissions
    )
    return user


def make_max_test_user(
    username=None,
    email='test@test.tld',
    password='t',
    is_staff=True,
    exclude_permissions=None
):
    """
    Create a test user with all permissions *except* the {exclude_permissions} ones.

    :param exclude_permissions: List of strings like: 'auth.import', 'app_label.codename', ...
    :return: User instance
    """
    permissions = Permission.objects.all()
    for permission in exclude_permissions:
        assert '.' in permission, f'Wrong format: {permission!r}'
        app_label, codename = permission.split('.')
        qs = Permission.objects.filter(
            codename=codename,
            content_type__app_label=app_label
        )
        assert qs.exists(), f'Not exists: {permission!r}'
        permissions = permissions.exclude(
            codename=codename,
            content_type__app_label=app_label
        )

    user = make_test_user(
        username=username,
        email=email,
        password=password,
        is_staff=is_staff,
        permissions=permissions
    )
    return user



import datetime
import decimal
import uuid
from typing import Any, Callable

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.functional import Promise


DJANGO_JSON_ENCODER_TYPES = (  # All types that DjangoJSONEncoder will convert
    datetime.datetime, datetime.date, datetime.time, datetime.timedelta,
    decimal.Decimal, uuid.UUID, Promise
)
JSON_SERIALIZEABLE_TYPES = (  # Without list, tuple, dict
    bool, str, int, float,
)
ALL_SERIALIZEABLE_TYPES = JSON_SERIALIZEABLE_TYPES + DJANGO_JSON_ENCODER_TYPES


def make_json_serializable(value: Any, convert_func: Callable = repr) -> Any:
    """
    Convert value to a JSON serializable value, with convert callback for special objects.
    Leave data types like datetime, UUID untouched, because DjangoJSONEncoder will handle this.
    """
    if value is None:
        return None
    elif isinstance(value, (set, list, tuple)):
        return [make_json_serializable(item, convert_func) for item in value]
    elif isinstance(value, dict):
        return {
            make_json_serializable(key, convert_func): make_json_serializable(value, convert_func)
            for key, value in value.items()
        }
    elif isinstance(value, ALL_SERIALIZEABLE_TYPES):
        return value

    return convert_func(value)


def to_json(
        value: Any,
        sort_keys=True,
        ensure_ascii=False,
        convert_func: Callable = repr,
        **json_kwargs
):
    """
    Convert value to JSON via make_json_serializable() and DjangoJSONEncoder()
    """
    value = make_json_serializable(value, convert_func=convert_func)
    value = DjangoJSONEncoder(
        sort_keys=sort_keys,
        ensure_ascii=ensure_ascii,
        **json_kwargs
    ).encode(value)
    return value
